import imp
import os

import pytest
from pyramid import testing
from webtest import TestApp
from sqlalchemy import create_engine
from pyramid_sacrud import CONFIG_MODELS
from sqlalchemy.orm import sessionmaker, scoped_session
from pyramid.threadlocal import get_current_registry
from sqlalchemy.ext.declarative import declarative_base

imp.load_source('pyramid_sacrud_example', 'example/pyramid_sacrud_example.py')

from pyramid_sacrud_example import (  # noqa
    Good,
    main,
    Tree,
    User,
    Group,
    models
)

DIRNAME = os.path.dirname(__file__)
DATABASE_FILE = os.path.join(DIRNAME, 'test.sqlite')
TEST_DATABASE_CONNECTION_STRING = 'sqlite:///%s' % DATABASE_FILE

Base = declarative_base()


class BaseFixture:

    def __init__(self):
        engine = create_engine(TEST_DATABASE_CONNECTION_STRING)
        self.engine = engine
        self.DBSession = scoped_session(sessionmaker(bind=self.engine))
        self.dbsession = self.DBSession(bind=self.engine)

    def setUp(self):
        # bind an individual Session to the connection
        self.create_db()

    def tearDown(self):
        self.drop_db()
        os.remove(DATABASE_FILE)
        self.dbsession.close()

    def drop_db(self):
        Base.metadata.drop_all(bind=self.engine)
        User.__table__.drop(self.engine)
        Group.__table__.drop(self.engine)
        Good.__table__.drop(self.engine)
        Tree.__table__.drop(self.engine)
        self.dbsession.commit()

    def create_db(self):
        Base.metadata.create_all(bind=self.engine)
        User.__table__.create(self.engine, checkfirst=True)
        Group.__table__.create(self.engine, checkfirst=True)
        Good.__table__.create(self.engine, checkfirst=True)
        Tree.__table__.create(self.engine, checkfirst=True)
        self.dbsession.commit()


@pytest.fixture(scope="function")
def unit(request):
    response = BaseFixture()
    response.setUp()
    response.request = testing.DummyRequest()
    response.config = testing.setUp(request=response.request)
    get_current_registry().settings[CONFIG_MODELS] = models
    get_current_registry().settings['sqlalchemy.url'] =\
        TEST_DATABASE_CONNECTION_STRING

    def fin():
        response.tearDown()
        testing.tearDown()

    request.addfinalizer(fin)
    return response


@pytest.fixture(scope="function")
def functional(request):
    settings = {
        'sqlalchemy.url': TEST_DATABASE_CONNECTION_STRING,
    }
    app = main({}, **settings)
    response = BaseFixture()
    response.setUp()
    response.app = TestApp(app)

    def fin():
        response.tearDown()

    request.addfinalizer(fin)
    return response



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

import colander
import sqlalchemy
from sqlalchemy import Column, String, Integer
from sacrud_deform import SacrudForm
from ps_alchemy.views import preprocessing_value
from sqlalchemy.ext.declarative import declarative_base


def test_preprocessing_value():

    Base = declarative_base()

    class User(Base):
        __tablename__ = 'users'
        id = Column(Integer, primary_key=True)
        foo = Column(String)
        bar = Column(Integer)

    form = SacrudForm(None, None, User)
    form.make_appstruct()
    assert preprocessing_value("foo", colander.null, form.schema) == ""
    assert preprocessing_value("foo", "", form.schema) == ""
    assert isinstance(
        sqlalchemy.sql.null(),
        type(preprocessing_value("bar", "", form.schema))
    )
    value = {'foo': 'bar'}
    assert preprocessing_value("foo", value, form.schema) == value



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

from pyramid.testing import DummyRequest
from ps_alchemy.paginator import get_paginator, get_current_page


def test_get_current_page():
    request = DummyRequest()
    page = get_current_page(request)
    assert page == 1
    request.GET['page'] = 5
    page = get_current_page(request)
    assert page == 5


def test_get_paginator():
    request = DummyRequest()
    paginator = get_paginator(request)
    assert paginator['items_per_page'] == 10
    assert paginator['page'] == 1

    request.GET['page'] = 100500
    paginator = get_paginator(request, 20)
    assert paginator['items_per_page'] == 20
    assert paginator['page'] == 100500



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.
from pyramid.events import ApplicationCreated
from pyramid_sacrud import CONFIG_MODELS
from pyramid_sacrud.resources import GroupResource
from sqlalchemy.ext.declarative.api import DeclarativeMeta

from .resources import ListResource


def models_preparing(app):
    """ Wrap all sqlalchemy model in settings.
    """
    settings = app.app.registry.settings
    models = settings[CONFIG_MODELS]

    def wrapper(resource, parent):
        if isinstance(resource, DeclarativeMeta):
            return ListResource(resource, parent=parent)
        return resource

    models = [(k, [wrapper(r, GroupResource(k, v)) for r in v])
              for k, v in models]
    settings[CONFIG_MODELS] = models


def includeme(config):
    config.add_subscriber(models_preparing, ApplicationCreated)

    config.include('pyramid_jinja2')
    config.add_jinja2_search_path("templates")
    config.scan('.views')

    # deform
    config.include('sacrud_deform')
    config.add_static_view('deform_static', 'deform:static')



import json
import logging

import deform
import colander
import peppercorn
import sqlalchemy
import transaction
from pyramid.view import view_config, view_defaults
from sacrud.common import pk_list_to_dict
from pyramid.compat import escape
from pyramid_sacrud import PYRAMID_SACRUD_VIEW
from pyramid.renderers import render_to_response
from sqlalchemy.orm.exc import NoResultFound
from paginate_sqlalchemy import SqlalchemyOrmPage
from pyramid.httpexceptions import HTTPFound, HTTPNotFound
from pyramid_sacrud.exceptions import SacrudException
from pyramid_sacrud.localization import _ps

from .paginator import get_paginator
from .resources import (
    ListResource,
    CreateResource,
    DeleteResource,
    MassDeleteResource
)


def preprocessing_value(key, value, form):
    for groups in form.children:
        for column in groups:
            if column.name == key:
                if not str(value).isdigit() and isinstance(
                        column.typ,
                        (colander.Int, colander.Integer,
                         colander.Float, colander.Decimal)):
                    value = sqlalchemy.sql.null()
                elif value is colander.null:
                    value = ""
                return value


@view_defaults(permission=PYRAMID_SACRUD_VIEW)
class CRUD(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def commit(self):
        try:
            self.context.dbsession.commit()
        except AssertionError:
            transaction.commit()

    def abort(self):
        try:
            self.context.dbsession.rollback()
        except AssertionError:
            transaction.abort()

    def list_view_response(self):
        return HTTPFound(
            location='/' + self.request.sacrud_prefix + '/' +
            self.request.resource_path(
                self.context.left_sibling_breadcrumb
            )
        )

    def flash_message(self, message, status="success"):
        if hasattr(self.request, 'session'):
            self.request.session.flash([message, status])


class Read(CRUD):

    @view_config(
        request_method='GET',
        context=ListResource,
        route_name=PYRAMID_SACRUD_VIEW,
    )
    def list_view(self):
        rows = self.context.crud.read()
        try:
            paginator_attr = get_paginator(
                self.request, self.context.items_per_page - 1
            )
        except ValueError:
            raise HTTPNotFound
        params = {
            'paginator': SqlalchemyOrmPage(rows, **paginator_attr)
        }
        return render_to_response(
            self.context.renderer, params, request=self.request
        )


class Create(CRUD):

    @view_config(
        request_method='GET',
        context=CreateResource,
        route_name=PYRAMID_SACRUD_VIEW
    )
    def edit_form_get_view(self):
        params = {'form': self.context.form(self.request).render()}
        return render_to_response(
            self.context.renderer, params, request=self.request
        )

    @view_config(
        request_method='POST',
        context=CreateResource,
        route_name=PYRAMID_SACRUD_VIEW
    )
    def edit_form_post_view(self):
        form = self.context.form(self.request)
        params = {'form': form.render()}

        def get_reponse(form=None):
            if form:
                params['form'] = form
            return render_to_response(
                self.context.renderer, params, request=self.request
            )

        if 'form.submitted' in self.request.params:
            controls = self.request.POST.items()
            pstruct = peppercorn.parse(controls)

            # Validate form
            try:
                deserialized = form.validate_pstruct(pstruct).values()
            except deform.ValidationFailure as e:
                return get_reponse(e.render())
            data = {k: preprocessing_value(k, v, form)  # TODO: optimize it
                    for d in deserialized
                    for k, v in d.items()}

            # Update object
            try:
                if self.context.obj:
                    obj = self.context.crud._add(self.context.obj, data)
                    flash_action = 'updated'
                else:
                    obj = self.context.crud.create(data)
                    flash_action = 'created'
                name = obj.__repr__()
                self.context.dbsession.flush()
            except SacrudException as e:
                self.flash_message(e.message, status=e.status)
                return get_reponse()
            except Exception as e:
                self.abort()
                logging.exception("Something awful happened!")
                raise e

            self.commit()

            # Make response
            self.flash_message(_ps(
                u"You ${action} object of ${name}",
                mapping={'action': flash_action, 'name': escape(name or '')}
            ))
            return self.list_view_response()
        return get_reponse()


class Delete(CRUD):

    @view_config(
        request_method='POST',
        context=DeleteResource,
        route_name=PYRAMID_SACRUD_VIEW
    )
    def delete_view(self):
        if not hasattr(self.context, 'obj'):
            raise HTTPNotFound
        self.context.dbsession.delete(self.context.obj)
        self.commit()
        self.flash_message(
            _ps("You have removed object of ${name}",
                mapping={'name': escape(str(self.context.obj) or '')}))
        return self.list_view_response()

    @view_config(
        request_method='POST',
        context=MassDeleteResource,
        route_name=PYRAMID_SACRUD_VIEW
    )
    def mass_delete_view(self):
        items_list = self.request.POST.getall('selected_item')
        primary_keys = [pk_list_to_dict(json.loads(item))
                        for item in items_list]
        objects = self.context.crud.read(*primary_keys)
        try:
            if hasattr(objects, 'delete'):
                object_names = [escape(x.__repr__() or '') for x in objects]
                objects.delete()
            else:
                object_names = [escape(objects.__repr__() or ''), ]
                self.request.dbsession.delete(objects)
        except (NoResultFound, KeyError):
            raise HTTPNotFound
        except SacrudException as e:
            self.flash_message(e.message, status=e.status)
        except Exception as e:
            transaction.abort()
            logging.exception("Something awful happened!")
            raise e
        transaction.commit()
        self.flash_message(_ps("You delete the following objects:"))
        [self.flash_message(obj) for obj in object_names]
        return self.list_view_response()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Functional tests for views
"""
from .conftest import User, Group, Tree


def is_table_exist(table_name, engine):
    return engine.dialect.has_table(engine.connect(), table_name)


def user_fixture(dbsession, ids=(1,)):
    for id in ids:
        user = User(id=id, name="some user %s" % id)
        dbsession.add(user)
    dbsession.commit()


class TestHome:

    def test_sa_home(self, functional):
        res = functional.app.get('/admin/', status=200)
        assert 'auth' in str(res.body)
        assert 'user' in str(res.body)
        assert 'goods' in str(res.body)


class TestCreate:

    def test_sa_create(self, functional):
        functional.app.get('/admin/auth/user/create/', status=200)

    def test_sa_create_table_not_exist(self, functional):
        functional.app.get('/admin/not_exist_table/create/', status=404)
        assert is_table_exist('not_exist_table', functional.engine) is False

    def test_sa_create_post(self, functional):
        post = [
            ('form.submitted', '1'),
            ('__formid__', 'deform'),
            ('__start__', ':mapping'),
            ('id', '1'),
            ('name', 'foo bar baz'),
            ('__end__', ':mapping')
        ]
        functional.app.post('/admin/auth/user/create/', post, status=302)
        user = functional.dbsession.query(User)\
            .order_by(User.id.desc()).first()
        assert user.name == 'foo bar baz'

    def test_sa_create_update_none_repr_obj(self, functional):
        post = [
            ('form.submitted', '1'),
            ('__formid__', 'deform'),
            ('__start__', ':mapping'),
            ('name', 'foo'),
            ('__end__', ':mapping')
        ]
        functional.app.post('/admin/Catalouge/groups/create/',
                            post, status=302)
        group = functional.dbsession.query(Group).count()
        assert group > 0

        post = [
            ('form.submitted', '1'),
            ('__formid__', 'deform'),
            ('__start__', ':mapping'),
            ('id', 2),
            ('__end__', ':mapping')
        ]
        functional.app.post('/admin/Catalouge/groups/update/id/1/',
                            post, status=302)
        group = functional.dbsession.query(Group).count()
        assert group > 0

# def test_sa_create_an_existing_obj(self, functional):
#     functional.drop_db()
#     functional.create_db()
#     post = [
#         ('form.submitted', '1'),
#         ('__formid__', 'deform'),
#         ('__start__', ':mapping'),
#         ('id', 1),
#         ('__end__', ':mapping')
#     ]
#     for x in range(10):
#         functional.app.post('/admin/Catalouge/groups/create/',
#                             post, status=302)
#     group = functional.dbsession.query(Group).count()
#     assert group > 0
#     with pytest.raises(IntegrityError):
#         functional.app.post('/admin/Catalouge/groups/create/',
#                             post, status=302)


class TestList:

    def test_sa_list(self, functional):
        functional.app.get('/admin/auth/user/', status=200)
        functional.app.get('/admin/Catalouge/groups/', status=200)

    def test_sa_update(self, functional):
        functional.dbsession.add(
            User(id=1, name='admin')
        )
        functional.dbsession.commit()
        functional.app.get('/admin/auth/user/update/id/1/name/admin',
                           status=200)

    def test_sa_update_not_found(self, functional):
        functional.app.get('/admin/auth/user/update/id/100500/name/foo',
                           status=404)

    def test_sa_update_bad_pk(self, functional):
        functional.app.get('/admin/auth/user/update/', status=404)
        functional.app.get('/admin/auth/user/update/id', status=404)
        functional.app.get('/admin/auth/user/update/not_pk/1', status=404)
        functional.app.get('/admin/auth/user/update/id/1/foo/2', status=404)

    def test_sa_update_table_not_exist(self, functional):
        functional.app.get('/admin/not_exist_table/update/id/100500',
                           status=404)
        assert is_table_exist('not_exist_table', functional.engine) is False

    def test_sa_update_post(self, functional):
        functional.dbsession.add(
            User(id=1, name='no_name')
        )
        functional.dbsession.commit()
        post = [
            ('form.submitted', '1'),
            ('__formid__', 'deform'),
            ('__start__', ':mapping'),
            ('sex', 'male'),
            ('name', 'foo bar baz'),
            ('__end__', ':mapping')
        ]
        functional.app.post('/admin/auth/user/update/id/1/name/no_name',
                            post, status=302)
        user = functional.dbsession.query(User).filter_by(id=1).one()
        assert user.name == 'foo bar baz'

    def test_sa_update_post_bad_data(self, functional):
        functional.dbsession.add(
            User(id=1, name='no_name')
        )
        functional.dbsession.commit()
        post = [
            ('form.submitted', '1'),
            ('__formid__', 'deform'),
            ('__start__', ':mapping'),
            ('id', 111),
            ('age', ''),
            ('sex', 'male'),
            ('name', 'some user 1'),
            ('__end__', ':mapping')
        ]
        functional.app.post('/admin/auth/user/update/id/1/name/no_name',
                            post, status=302)
        user = functional.dbsession.query(User).filter_by(id=111).one()
        assert user.name == 'some user 1'


class TestDelete:

    def test_delete_user(self, functional):
        functional.dbsession.add(
            User(id=1, name='foo')
        )
        functional.dbsession.commit()
        functional.app.post('/admin/auth/user/delete/id/1/name/foo',
                            status=302)
        user = functional.dbsession.query(User).filter_by(id=1).all()
        assert len(user) == 0

    def test_delete_not_found_user(self, functional):
        functional.app.post('/admin/auth/user/delete/id/100500', status=404)
        user = functional.dbsession.query(User).filter_by(id=100500).all()
        assert len(user) == 0

    def test_sa_delete_bad_pk(self, functional):
        functional.dbsession.add(
            User(id=1, name='foo')
        )
        functional.dbsession.commit()
        functional.app.get('/admin/auth/user/delete/', status=404)
        functional.app.get('/admin/auth/user/delete/id', status=404)
        functional.app.get('/admin/auth/user/delete/not_pk/1', status=404)
        functional.app.get('/admin/auth/user/delete/id/1/foo/2', status=404)

    def test_delete_not_exist_table(self, functional):
        functional.app.get('/admin/not_exist_table/delete/id/1', status=404)
        assert is_table_exist('not_exist_table', functional.engine) is False

    def test_sa_list_delete_actions(self, functional):
        ids = range(1, 100)
        user_fixture(functional.dbsession, ids=ids)
        deleted_ids = range(11, 25)
        items_list = [u'["id", %s, "name", "some user %s"]' % (id, id)
                      for id in deleted_ids]
        functional.app.post('/admin/auth/user/mass_action/delete/',
                            {'selected_item': items_list},
                            status=302)
        count = functional.dbsession.query(User).filter(
            User.id.in_(deleted_ids)
        ).count()
        assert count == 0
        count = functional.dbsession.query(User).filter(
            User.id.in_(ids)
        ).count()
        assert count == len(ids) - len(deleted_ids)

    def test_sa_list_delete_cascade_actions(self, functional):
        functional.dbsession.add(Tree(id=1))
        functional.dbsession.add(Tree(id=2, parent_id=1))
        functional.dbsession.add(Tree(id=3, parent_id=2))
        functional.dbsession.add(Tree(id=4, parent_id=1))
        functional.dbsession.add(Tree(id=5))
        functional.dbsession.commit()

        deleted_ids = (1, 3, 4)
        items_list = [u'["id", %s]' % id for id in deleted_ids]
        functional.app.post('/admin/foo/tree/mass_action/delete/',
                            {'selected_item': items_list},
                            status=302)
        objects = [x.id for x in functional.dbsession.query(Tree).all()]
        assert 1 not in objects
        assert 2 in objects
        assert 3 not in objects
        assert 4 not in objects
        assert 5 in objects



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Webhelpers paginator
"""
from paginate import Page, make_html_tag


def paginate_link_tag(item):
    """
    Create an A-HREF tag that points to another page usable in paginate.
    """
    a_tag = Page.default_link_tag(item)
    if item['type'] == 'current_page':
        return make_html_tag('li', a_tag, **{'class': 'blue white-text'})
    return make_html_tag('li', a_tag)


def get_current_page(request):
    return int(request.params.get('page', 1))


def get_paginator(request, items_per_page=10):
    return {
        "items_per_page": items_per_page,
        "page": get_current_page(request),
        "url_maker": lambda p: request.path_url + "?page=%s" % p,
        "link_tag": paginate_link_tag
    }



import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

this = os.path.dirname(os.path.realpath(__file__))


def read(name):
    with open(os.path.join(this, name)) as f:
        return f.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = [
            '--cov=ps_alchemy',
            '-s'
        ]

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.verbose = True
        self.test_suite = 'tests'

    def run_tests(self):
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(
    name='ps_alchemy',
    version='0.0.2',
    url='http://github.com/sacrud/ps_alchemy/',
    author='Svintsov Dmitry',
    author_email='sacrud@uralbash.ru',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    test_suite="pytest",
    cmdclass={'test': PyTest},
    license="MIT",
    description='SQLAlchemy provider for pyramid_sacrud.',
    long_description=read('README.rst'),
    install_requires=read('requirements.txt'),
    tests_require=read('requirements.txt') + read('requirements-test.txt'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Framework :: Pyramid ",
        "Topic :: Internet",
        "Topic :: Database",
    ],
)
